package org.processmining.dataawarecnetminer.converter;

import java.util.Map;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.processmining.dataawarecnetminer.extension.bindings.Bindings;
import org.processmining.models.causalgraph.Relation;
import org.processmining.models.causalgraph.SimpleCausalGraph;
import org.processmining.models.cnet.CNetBinding;
import org.processmining.models.cnet.CNetBinding.Type;
import org.processmining.models.cnet.CNetNode;
import org.processmining.models.cnet.CausalNet;
import org.processmining.models.dcnet.DCNet;
import org.processmining.models.dcnet.DataAwareCausalNet;

import com.google.common.base.Function;
import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableMap.Builder;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Iterables;
import com.google.common.collect.Multiset;

public class CausalToCNet {

	private static final class CausalNetImpl implements CausalNet {

		private final BiMap<XEventClass, CNetNode> classToNode;

		private final XEventClass startTask;
		private final XEventClass endTask;

		private final ImmutableMap<CNetNode, Set<CNetNode>> successors;
		private final ImmutableMap<CNetNode, Set<CNetNode>> predecessors;

		private final ImmutableMap<CNetNode, Set<CNetBinding>> outputBindings;
		private final ImmutableMap<CNetNode, Set<CNetBinding>> inputBindings;

		private CausalNetImpl(SimpleCausalGraph causalGraph, BiMap<XEventClass, CNetNode> classToNode, Bindings bindings,
				XEventClass endTask, XEventClass startTask) {
			this.classToNode = classToNode;
			this.endTask = endTask;
			this.startTask = startTask;

			Builder<CNetNode, Set<CNetNode>> predecessorsBuilder = ImmutableMap.builder();
			Builder<CNetNode, Set<CNetNode>> successorsBuilder = ImmutableMap.builder();

			Builder<CNetNode, Set<CNetBinding>> outputBuilder = ImmutableMap.builder();
			Builder<CNetNode, Set<CNetBinding>> inputBuilder = ImmutableMap.builder();

			for (XEventClass activity : causalGraph.getSetActivities()) {
				CNetNode node = classToNode.get(activity);
				if (node != null) {
					predecessorsBuilder.put(node, ImmutableSet.copyOf(transform(causalGraph.getPreset(activity))));
					successorsBuilder.put(node, ImmutableSet.copyOf(transform(causalGraph.getPostset(activity))));

					Multiset<Set<XEventClass>> outputSets = bindings.getOutputSets(activity);
					if (outputSets != null) {
						outputBuilder.put(node, ImmutableSet.copyOf(transform(Type.OUTPUT, node, outputSets.elementSet())));
					}

					Multiset<Set<XEventClass>> inputSets = bindings.getInputSets(activity);
					if (inputSets != null) {
						inputBuilder.put(node, ImmutableSet.copyOf(transform(Type.INPUT, node, inputSets.elementSet())));	
					}					
				}
			}
			predecessors = predecessorsBuilder.build();
			successors = successorsBuilder.build();

			inputBindings = inputBuilder.build();
			outputBindings = outputBuilder.build();
		}

		public Set<CNetNode> getSuccessors(CNetNode node) {
			return successors.get(node);
		}

		public Set<CNetNode> getPredecessors(CNetNode node) {
			return predecessors.get(node);
		}

		public CNetNode getStartNode() {
			return classToNode.get(startTask);
		}

		public CNetNode getEndNode() {
			return classToNode.get(endTask);
		}

		public Set<CNetNode> getNodes() {
			return classToNode.values();
		}

		public Set<CNetBinding> getOutputBindings(CNetNode node) {
			return outputBindings.get(node);
		}

		public Set<CNetBinding> getInputBindings(CNetNode node) {
			return inputBindings.get(node);
		}

		private Iterable<CNetNode> transform(Set<XEventClass> tasks) {
			return Iterables.transform(tasks, new Function<XEventClass, CNetNode>() {

				public CNetNode apply(XEventClass clazz) {
					return classToNode.get(clazz);
				}
			});
		}

		private Iterable<CNetBinding> transform(final Type type, final CNetNode node, Set<Set<XEventClass>> bindings) {
			return Iterables.transform(bindings, new Function<Set<XEventClass>, CNetBinding>() {

				public CNetBinding apply(Set<XEventClass> binding) {
					return new CNetBinding(type, node, ImmutableSet.copyOf(transform(binding)));
				}
			});
		}

		public String getLabel() {
			return "";
		}

		public String toString() {
			return "Successors: " + successors.toString() + "\nPredecessors: " + predecessors.toString();
		}
		

	}

	public static DataAwareCausalNet convert(final SimpleCausalGraph causalGraph, final Bindings bindings,
			final XEventClass startTask, final XEventClass endTask) {
		final BiMap<XEventClass, CNetNode> classToNode = HashBiMap.create();
		for (Relation relation : causalGraph.getCausalRelations()) {
			createIfNonExisting(classToNode, relation.getSource());
			createIfNonExisting(classToNode, relation.getTarget());
		}
		return new DCNet(new CausalNetImpl(causalGraph, classToNode, bindings, endTask, startTask));
	}

	private static CNetNode createIfNonExisting(final Map<XEventClass, CNetNode> classToNode, XEventClass eventClass) {
		CNetNode node = classToNode.get(eventClass);
		if (node == null) {
			node = new CNetNode(eventClass.getId());
			classToNode.put(eventClass, node);
		}
		return node;
	}

}
