package org.processmining.dataawarecnetminer.interactive;

import java.text.MessageFormat;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.dataawarecnetminer.model.DataRelationStorage;
import org.processmining.dataawarecnetminer.model.EventRelationStorage;
import org.processmining.dataawarecnetminer.util.DataDiscoveryUtil;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.log.utils.XUtils;
import org.processmining.plugins.utils.ProvidedObjectHelper;

import com.google.common.base.Stopwatch;
import com.google.common.collect.ImmutableSet;

/**
 * Interactive plugin with the data-aware Heuristic Miner settings
 * 
 * @author F. Mannhardt
 *
 */
public final class InteractiveDataAwareCausalMinerPlugin {

	@Plugin(name = "Interactive Data-aware Heuristic Miner (iDHM)", parameterLabels = { "Event log" }, //
			keywords = { "discover", "heuristric", "interactive", "data", "decision" }, categories = {
					PluginCategory.Discovery, PluginCategory.ConformanceChecking }, // 
			help = "The interactive Data-aware Heuristic Miner (iDHM) is a tool that enables interactive exploration of heuristic process discovery approached with built-in conformance checking. "
					+ "The iDHM uses uses a plug-in architecture to orchestrate several heuristics that are used during the process discovery. The employed heuristics for the different phases of heuristic discovery, i.e, (A) mining dependencies, (B) adding conditional dependencies, (C) mining split- and join gateways, and (D) checking conformance, can be freely combined. "
					+ "Relevant papers are published as: F. Mannhardt, M. de Leoni, H. A. Reijers, W. M. P. van der Aalst (To Appear). Data-driven Process Discovery - Revealing Conditional Infrequent Behavior from Event Logs. In CAiSE 2017"
					+ "Documentation is available at: https://fmannhardt.de/blog/data-aware-heuristic-miner-dhm", //
			level = PluginLevel.PeerReviewed, returnLabels = {
					"Interactive Data-aware Heuristic Miner" }, returnTypes = {
							InteractiveDataAwareCausalMiner.class }, userAccessible = true, mostSignificantResult = 1)
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public InteractiveDataAwareCausalMiner dataAwareHeuristicMiner(UIPluginContext context, XLog log)
			throws UserCancelledException {
		context.getProgress().setMaximum(3);

		XEventClassifier classifier = ProMUIHelper.queryForObject(context, "Select the event classifier to be used",
				XUtils.getStandardAndLogDefinedEventClassifiers(log));

		Map<String, Class<?>> attributeTypes = DataDiscoveryUtil.getRelevantAttributeTypes(log);

		context.log("Computing log relations ...");
		context.getProgress().setValue(1);

		Stopwatch watch = Stopwatch.createStarted();
		ExecutorService executorService = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
		try {
			EventRelationStorage eventRelationStorage = EventRelationStorage.Factory.createEventRelations(log,
					classifier, executorService);
			DataRelationStorage dataRelations = DataRelationStorage.Factory.createDataRelations(eventRelationStorage,
					attributeTypes, ImmutableSet.<String>of(), executorService);

			watch.stop();
			context.getProgress().setValue(2);
			context.log(MessageFormat.format("Finished in {0,number,#}ms", watch.elapsed(TimeUnit.MILLISECONDS)));

			context.getProgress().setValue(3);

			context.getFutureResult(0).setLabel("Interactive Data-aware Heuristic Miner for "
					+ ProvidedObjectHelper.getProvidedObjectLabel(context, log));

			return new InteractiveDataAwareCausalMiner(eventRelationStorage, dataRelations);

		} finally {
			executorService.shutdown();
		}
	}

}
