package org.processmining.dataawarecnetminer.mining.data;

import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.processmining.dataawarecnetminer.common.MinerContext;
import org.processmining.dataawarecnetminer.extension.conditionaldependencies.ConditionalDependencyHeuristicConfig;
import org.processmining.dataawarecnetminer.model.DataAwareCausalGraph;
import org.processmining.dataawarecnetminer.model.DataRelation;
import org.processmining.dataawarecnetminer.model.DataRelationStorage;
import org.processmining.dataawarecnetminer.model.DependencyRelation;
import org.processmining.dataawarecnetminer.model.EventRelationStorage;
import org.processmining.datadiscovery.RuleDiscoveryException;
import org.processmining.models.causalgraph.Relation;
import org.processmining.models.causalgraph.SimpleCausalGraph;

import com.google.common.base.Predicate;
import com.google.common.collect.Sets;

public final class DataAwareCausalGraphMiner {

	private final class DCGraphWithArtificalStartEnd implements DataAwareCausalGraph {

		private final DataAwareCausalGraph causalGraph;

		private DCGraphWithArtificalStartEnd(DataAwareCausalGraph causalGraph) {
			this.causalGraph = causalGraph;
		}

		public Set<XEventClass> getPreset(XEventClass activity) {
			return causalGraph.getPreset(activity);
		}

		public Set<XEventClass> getPostset(XEventClass activity) {
			return causalGraph.getPostset(activity);
		}

		public Set<Relation> getCausalRelations() {
			return Sets.filter(causalGraph.getCausalRelations(), new Predicate<Relation>() {

				public boolean apply(Relation relation) {
					return !relation.getSource().equals(eventRelations.getStartEventClass())
							&& !relation.getTarget().equals(eventRelations.getEndEventClass());
				}
			});
		}

		public Set<XEventClass> getSetActivities() {
			return Sets.filter(causalGraph.getSetActivities(), new Predicate<XEventClass>() {

				public boolean apply(XEventClass eventClass) {
					return !eventClass.equals(eventRelations.getStartEventClass())
							&& !eventClass.equals(eventRelations.getEndEventClass());
				}
			});
		}

		public DependencyRelation getDependencyRelation(Relation relation) {
			return causalGraph.getDependencyRelation(relation);
		}

		public DataRelation getDataRelation(Relation relation) {
			return causalGraph.getDataRelation(relation);
		}
	}

	private EventRelationStorage eventRelations;
	private DataRelationStorage dataRelations;

	private boolean includeArtificialStartEnd = true;
	private DataAwareCausalGraphBuilder graphBuilder;

	public DataAwareCausalGraphMiner(EventRelationStorage eventRelations, DataRelationStorage dataRelations) {
		this.eventRelations = eventRelations;
		this.dataRelations = dataRelations;
		this.graphBuilder = new DataAwareCausalGraphBuilder(eventRelations, dataRelations);
	}

	public DataAwareCausalGraph mineCausalGraph(SimpleCausalGraph initialCausalGraph) throws RuleDiscoveryException {
		return addStartEndIfNecessary(graphBuilder.build(initialCausalGraph));
	}

	public DataAwareCausalGraph mineCausalGraph(MinerContext context, SimpleCausalGraph initialCausalGraph)
			throws RuleDiscoveryException {
		return addStartEndIfNecessary(graphBuilder.build(initialCausalGraph));
	}

	private DataAwareCausalGraph addStartEndIfNecessary(final DataAwareCausalGraph causalGraph) {
		if (isIncludeArtificialStartEnd()) {
			return causalGraph;
		} else {
			return new DCGraphWithArtificalStartEnd(causalGraph);
		}
	}

	private boolean isIncludeArtificialStartEnd() {
		return includeArtificialStartEnd;
	}

	public void setDataDiscoveryAttributes(Set<String> dataDiscoveryAttributes) {
		dataRelations.setConsideredAttributes(dataDiscoveryAttributes);
	}

	public Set<String> getDataDiscoveryAttributes() {
		return dataRelations.getConsideredAttributes();
	}

	public ConditionalDependencyHeuristicConfig getConfig() {
		return graphBuilder.getConfig();
	}

	public void setConfig(ConditionalDependencyHeuristicConfig config) {
		graphBuilder.setConfig(config);
	}

}