package org.processmining.dataawarecnetminer.mining.decision;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.impl.XEventImpl;
import org.processmining.dataawarecnetminer.extension.decision.DecisionHeuristicConfig;
import org.processmining.dataawarecnetminer.model.DataRelationStorage;
import org.processmining.dataawarecnetminer.model.DiscoveredDCNet;
import org.processmining.dataawarecnetminer.model.DiscoveredDataAwareCausalNet;
import org.processmining.dataawarecnetminer.model.EventRelationStorage;
import org.processmining.dataawarecnetminer.model.FrequencyAwareCausalNet;
import org.processmining.dataawarecnetminer.model.FrequencyAwareDataCausalNet;
import org.processmining.dataawarecnetminer.parallel.Command;
import org.processmining.dataawarecnetminer.parallel.TaskExecutor;
import org.processmining.dataawarecnetminer.util.DataDiscoveryUtil;
import org.processmining.datadiscovery.BasicDecisionTreeImpl;
import org.processmining.datadiscovery.DecisionTreeConfig;
import org.processmining.datadiscovery.ProjectedEvent;
import org.processmining.datadiscovery.ProjectedLog;
import org.processmining.datadiscovery.ProjectedTrace;
import org.processmining.datadiscovery.RuleDiscovery;
import org.processmining.datadiscovery.RuleDiscovery.Rule;
import org.processmining.datadiscovery.RuleDiscoveryException;
import org.processmining.datadiscovery.estimators.FunctionEstimation;
import org.processmining.datadiscovery.estimators.Type;
import org.processmining.log.utils.XUtils;
import org.processmining.models.cnet.CNetBinding;
import org.processmining.models.cnet.CNetNode;
import org.processmining.models.cnet.CausalNet;

import com.google.common.collect.ConcurrentHashMultiset;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.ImmutableSet.Builder;
import com.google.common.collect.Multiset;

public class DataAwareCausalNetMiner {

	private static final class FrequencyAwareDataCausalNetImpl implements FrequencyAwareDataCausalNet {
		private final FrequencyAwareCausalNet frequencyAwareCausalNet;
		private final DiscoveredDCNet dataCausalNet;

		private FrequencyAwareDataCausalNetImpl(FrequencyAwareCausalNet frequencyAwareCausalNet,
				DiscoveredDCNet dataCausalNet) {
			this.frequencyAwareCausalNet = frequencyAwareCausalNet;
			this.dataCausalNet = dataCausalNet;
		}

		public int getNodeFrequency(CNetNode node) {
			return frequencyAwareCausalNet.getNodeFrequency(node);
		}

		public int getInputFrequency(CNetNode node) {
			return frequencyAwareCausalNet.getInputFrequency(node);
		}

		public int getOutputFrequency(CNetNode node) {
			return frequencyAwareCausalNet.getOutputFrequency(node);
		}

		public int getBindingFrequency(CNetBinding binding) {
			return frequencyAwareCausalNet.getBindingFrequency(binding);
		}

		public DiscoveredDataAwareCausalNet getDCNet() {
			return dataCausalNet;
		}

		public int getNodeSkippedFrequency(CNetNode node) {
			return frequencyAwareCausalNet.getNodeSkippedFrequency(node);
		}

		public int getNodeInsertedFrequency(CNetNode node) {
			return frequencyAwareCausalNet.getNodeInsertedFrequency(node);
		}

		public int getIncorrectVariableFrequency(CNetNode node, String variable) {
			return frequencyAwareCausalNet.getIncorrectVariableFrequency(node, variable);
		}

		public int getCorrectVariableFrequency(CNetNode node, String variable) {
			return frequencyAwareCausalNet.getCorrectVariableFrequency(node, variable);
		}

		public int getMissingVariableFrequency(CNetNode node, String variable) {
			return frequencyAwareCausalNet.getMissingVariableFrequency(node, variable);
		}
	}

	private static final class FrequencyAwareCausalNetPassthrough implements FrequencyAwareCausalNet {
		private final CausalNet causalNet;

		private FrequencyAwareCausalNetPassthrough(CausalNet causalNet) {
			this.causalNet = causalNet;
		}

		public int getBindingFrequency(CNetBinding binding) {
			return -1;
		}

		public int getInputFrequency(CNetNode node) {
			return -1;
		}

		public int getOutputFrequency(CNetNode node) {
			return -1;
		}

		public int getNodeFrequency(CNetNode node) {
			return -1;
		}

		public CausalNet getCNet() {
			return causalNet;
		}

		public int getNodeSkippedFrequency(CNetNode node) {
			return 0;
		}

		public int getNodeInsertedFrequency(CNetNode node) {
			return 0;
		}

		public int getCorrectVariableFrequency(CNetNode node, String variable) {
			return 0;
		}

		public int getMissingVariableFrequency(CNetNode node, String variable) {
			return 0;
		}

		public int getIncorrectVariableFrequency(CNetNode node, String variable) {
			return 0;
		}
	}

	final static boolean DEBUG = true;

	private EventRelationStorage eventRelations;
	private DataRelationStorage dataRelations;

	// Configuration
	private DecisionHeuristicConfig config = new DecisionHeuristicConfig();

	public DataAwareCausalNetMiner(EventRelationStorage eventRelations, DataRelationStorage dataRelations) {
		this.eventRelations = eventRelations;
		this.dataRelations = dataRelations;
	}

	private Set<String> convertToIds(Set<CNetNode> tasks) {
		Builder<String> inputTasksIds = ImmutableSet.<String>builder();
		for (CNetNode task : tasks) {
			inputTasksIds.add(task.getLabel());
		}
		return inputTasksIds.build();
	}

	public FrequencyAwareDataCausalNet mineDataAwareCausalNet(final CausalNet causalNet) throws RuleDiscoveryException {
		return mineDataAwareCausalNet(new FrequencyAwareCausalNetPassthrough(causalNet));
	}

	public FrequencyAwareDataCausalNet mineDataAwareCausalNet(FrequencyAwareCausalNet frequencyAwareCausalNet)
			throws RuleDiscoveryException {
		ExecutorService executor = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
		try {
			return mineDataAwareCausalNet(frequencyAwareCausalNet, executor);
		} finally {
			executor.shutdown();
		}
	}

	public FrequencyAwareDataCausalNet mineDataAwareCausalNet(final FrequencyAwareCausalNet frequencyAwareCausalNet,
			ExecutorService executor) throws RuleDiscoveryException {

		CausalNet causalNet = frequencyAwareCausalNet.getCNet();

		if (!getConfig().getDataDiscoveryAttributes().isEmpty()) {
			XEvent startEvent = new XEventImpl();
			if (causalNet.getStartNode() != null) {
				XUtils.assignConceptName(startEvent, causalNet.getStartNode().getLabel());
			} else {
				XUtils.assignConceptName(startEvent, EventRelationStorage.ARTIFICIAL_START);
			}
			XEvent endEvent = new XEventImpl();
			if (causalNet.getEndNode() != null) {
				XUtils.assignConceptName(endEvent, causalNet.getEndNode().getLabel());
			} else {
				XUtils.assignConceptName(endEvent, EventRelationStorage.ARTIFICIAL_END);
			}

			final DiscoveredDCNet dataCausalNet = discoverDecisionRules(executor, causalNet, startEvent, endEvent);
			addWriteOperations(executor, causalNet, dataCausalNet, startEvent, endEvent);

			return new FrequencyAwareDataCausalNetImpl(frequencyAwareCausalNet, dataCausalNet);
		} else {
			return new FrequencyAwareDataCausalNetImpl(frequencyAwareCausalNet, new DiscoveredDCNet(causalNet));
		}

	}

	private Map<CNetNode, Set<String>> buildPredecessorMap(CausalNet causalNet) {
		Map<CNetNode, Set<String>> predecessorIds = new HashMap<>();
		for (CNetNode node : causalNet.getNodes()) {
			predecessorIds.put(node, convertToIds(causalNet.getPredecessors(node)));
		}
		return predecessorIds;
	}

	private DiscoveredDCNet discoverDecisionRules(ExecutorService executor, CausalNet causalNet, XEvent startEvent,
			XEvent endEvent) throws RuleDiscoveryException {

		final DiscoveredDCNet dataCausalNet = new DiscoveredDCNet(causalNet);

		final Map<CNetNode, Set<String>> predecessorIds = buildPredecessorMap(causalNet);

		List<Callable<Set<GuardedBinding>>> callables = new ArrayList<>();
		for (CNetNode node : causalNet.getNodes()) {
			Set<CNetBinding> outputBindings = causalNet.getOutputBindings(node);
			if (outputBindings.size() > 1) {
				ActivityMapper activityMapper = createActivityMapper(causalNet, node, outputBindings, predecessorIds,
						eventRelations.getClassifier());

				callables.add(new DefaultGuardedBindingDiscovery(eventRelations.getAllTraces(), node, outputBindings,
						startEvent, endEvent, activityMapper, getConfig()) {

					protected RuleDiscovery<? extends Rule> getRuleDiscoveryMethod(ProjectedLog projectedLog) {
						return createRuleDiscovery(projectedLog, getConfig().getDataDiscoveryConfig(),
								DataDiscoveryUtil.viewSelectedOnly(dataRelations.getAttributeTypes(),
										getConfig().getDataDiscoveryAttributes()),
								dataRelations.getLiteralValues(), eventRelations.countTraces());
					}

				});
			}
		}

		if (DEBUG) {
			System.out.println("relation,quality,rule");
		}

		try {
			for (Future<Set<GuardedBinding>> future : executor.invokeAll(callables)) {
				Set<GuardedBinding> guardedBindings = future.get();
				for (GuardedBinding guardedBinding : guardedBindings) {
					FunctionEstimation estimation = guardedBinding.getFunctionEstimation();
					if (!estimation.getExpression().isTrue()) {
						if (DEBUG) {
							System.out.println(MessageFormat.format("\"{0}\",{1},\"{2}\"", guardedBinding.getBinding(),
									estimation.getExpression(), estimation.getQualityMeasure()));
						}
						dataCausalNet.setGuardExpression(guardedBinding.getBinding(), estimation.getExpression());
						dataCausalNet.setEstimation(guardedBinding.getBinding(), estimation);
						for (String variable : estimation.getExpression().getNormalVariables()) {
							Type type = dataRelations.getAttributeTypes().get(variable);
							if (type != null) {
								dataCausalNet.addVariable(variable, type.getJavaType());
							} else {
								throw new RuleDiscoveryException("Unknown variable " + variable
										+ ". Available variables are " + dataRelations.getAttributeTypes().keySet());
							}
						}
					}
				}
			}
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
		} catch (ExecutionException e) {
			throw new RuleDiscoveryException(e.getCause());
		}
		return dataCausalNet;
	}

	protected RuleDiscovery<? extends Rule> createRuleDiscovery(ProjectedLog projectedLog,
			DecisionTreeConfig discoveryConfig, Map<String, Type> attributesTypes,
			Map<String, Set<String>> literalValues, int estimateSize) {
		return new BasicDecisionTreeImpl(discoveryConfig, projectedLog, attributesTypes, literalValues, estimateSize);
	}

	protected ActivityMapper createActivityMapper(final CausalNet causalNet, final CNetNode node,
			final Set<CNetBinding> bindings, final Map<CNetNode, Set<String>> predecessorMap,
			final XEventClassifier classifier) {
		return new NearestActivityImplementation(node, causalNet, predecessorMap, bindings, classifier);
	}

	private void addWriteOperations(ExecutorService executor, final CausalNet causalNet,
			final DiscoveredDCNet dataCausalNet, XEvent startEvent, XEvent endEvent) {
		ProjectedLog projectedLog = new ProjectedLogImpl(eventRelations.getAllTraces(), new ActivityMapper() {

			public Object getActivity(XEvent currentEvent, XEvent previousEvent, List<XEvent> subTraceFromCurrent) {
				return XUtils.getConceptName(currentEvent);
			}
		}, startEvent, endEvent, getConfig().getDataDiscoveryAttributes());

		final Map<CNetNode, Multiset<String>> variablesWritten = new HashMap<>();
		final Multiset<CNetNode> nodeCounts = ConcurrentHashMultiset.<CNetNode>create();
		for (CNetNode node : causalNet.getNodes()) {
			variablesWritten.put(node, ConcurrentHashMultiset.<String>create());
		}
		final Map<String, CNetNode> nameToNode = createNodeNameLookup(causalNet);

		try {
			TaskExecutor.Factory.getInstance().execute(executor, projectedLog, new Command<ProjectedTrace>() {

				public void execute(ProjectedTrace trace) {
					for (ProjectedEvent event : trace) {
						CNetNode node = nameToNode.get(event.getActivity());
						if (node != null) {
							nodeCounts.add(node);
							for (String attribute : dataCausalNet.getVariables()) {
								if (event.getAttributeValue(attribute) != null) {
									variablesWritten.get(node).add(attribute);
								}
							}
						}
					}
				}

			});
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
		}

		for (CNetNode node : causalNet.getNodes()) {
			int nodeCount = nodeCounts.count(node);
			for (String variable : dataCausalNet.getVariables()) {
				int varCount = variablesWritten.get(node).count(variable);
				double varFrequency = varCount / (double) nodeCount;
				if (varFrequency > config.getWriteThreshold()) {
					dataCausalNet.addWrite(node, variable);
				}
			}
		}

	}

	private Map<String, CNetNode> createNodeNameLookup(CausalNet cNet) {
		final Map<String, CNetNode> nameToNode = new HashMap<>();
		for (CNetNode node : cNet.getNodes()) {
			nameToNode.put(node.getLabel(), node);
		}
		return nameToNode;
	}

	public DecisionHeuristicConfig getConfig() {
		return config;
	}

	public void setConfig(DecisionHeuristicConfig config) {
		this.config = config;
	}

}