package org.processmining.dataawarecnetminer.mining.decision;

import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XEvent;
import org.processmining.models.cnet.CNetBinding;
import org.processmining.models.cnet.CNetNode;
import org.processmining.models.cnet.CausalNet;

final class NearestActivityImplementation implements ActivityMapper {
	private final CNetNode node;
	private final CausalNet causalNet;
	private final Map<CNetNode, Set<String>> predecessorMap;
	private final Set<CNetBinding> bindings;
	private final XEventClassifier classifier;

	NearestActivityImplementation(CNetNode node, CausalNet causalNet,
			Map<CNetNode, Set<String>> predecessorMap, Set<CNetBinding> bindings, XEventClassifier classifier) {
		this.node = node;
		this.causalNet = causalNet;
		this.predecessorMap = predecessorMap;
		this.bindings = bindings;
		this.classifier = classifier;
	}

	public Object getActivity(XEvent currentEvent, XEvent previousEvent, List<XEvent> subTraceFromCurrent) {
		if (previousEvent != null && isSameActivity(node, previousEvent)) {
			return determineBinding(subTraceFromCurrent);
		} else {
			return null;
		}
	}

	private CNetBinding determineBinding(List<XEvent> remainingTrace) {

		Set<CNetNode> activatedOutputTasks = getActivatedOutputTasks(remainingTrace, node);
		if (!activatedOutputTasks.isEmpty()) {
			CNetBinding binding = new CNetBinding(CNetBinding.Type.OUTPUT, node, activatedOutputTasks);
			if (bindings.contains(binding)) {
				return binding;
			}
		}
		return null;
	}

	private Set<CNetNode> getActivatedOutputTasks(List<XEvent> remainingOfTheTrace, CNetNode currentTask) {
		Set<CNetNode> activatedTasks = new HashSet<>();
		Set<CNetNode> outputTasks = causalNet.getSuccessors(currentTask);

		// for each successor in the causal net graph
		for (CNetNode outputTask : outputTasks) {

			// look whether there is no other possibly activating task between task and outputTask
			Set<String> inputTasksIds = predecessorMap.get(outputTask);

			if (isNearestCandidateTask(remainingOfTheTrace, outputTask.getLabel(), inputTasksIds)) {
				// this outputTask was activated by currentTask as there is no other task from the inputSet in between and the outputTask is present
				activatedTasks.add(outputTask);
			}

		}
		return activatedTasks;
	}

	private boolean isNearestCandidateTask(List<XEvent> remainingOfTheTrace, String taskToFind,
			Set<String> candidateSet) {

		for (XEvent candidateEvent : remainingOfTheTrace) {
			String candidate = classifier.getClassIdentity(candidateEvent); //TODO mapping
			if (candidate.equals(taskToFind)) {
				return true; // we found the task without any candidate in between
			} else if (candidateSet.contains(candidate)) {
				// found another candidate beforehand
				return false;
			}
		}

		return taskToFind.equals(causalNet.getEndNode().getLabel())
				|| taskToFind.equals(causalNet.getStartNode().getLabel()); // the trace does not contain the taskToFind or we activated the endTask or the startTask
	}

	public boolean isSameActivity(CNetNode node, XEvent event) {
		return classifier.getClassIdentity(event).equals(node.getLabel());
	}

}