package org.processmining.dataawarecnetminer.model;

import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.processmining.dataawarecnetminer.parallel.Command;
import org.processmining.dataawarecnetminer.parallel.TaskExecutor;
import org.processmining.datadiscovery.estimators.Type;
import org.processmining.datadiscovery.estimators.weka.WekaUtil;
import org.processmining.datapetrinets.exception.VariableNotFoundException;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.datapetrinets.expression.VariableProvider;
import org.processmining.log.utils.XUtils;
import org.processmining.models.causalgraph.Relation;

import com.google.common.base.Function;
import com.google.common.collect.ConcurrentHashMultiset;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.ImmutableSortedMap;
import com.google.common.collect.Iterators;
import com.google.common.collect.Multiset;
import com.google.common.collect.Multiset.Entry;
import com.google.common.collect.PeekingIterator;
import com.google.common.collect.Sets;

public final class DataRelationStorageImpl implements DataRelationStorage {

	private interface AttributedEvent {

		XEventClass getEventClass();

		Collection<XAttribute> getAttributes();

	}

	private final class AttributedEventTransformation implements Function<XEvent, AttributedEvent> {
		public AttributedEvent apply(XEvent event) {
			final XEventClass eventClass = eventStore.getEventClass(event);
			final Collection<XAttribute> attributes = event.getAttributes().values();
			return new AttributedEvent() {

				public XEventClass getEventClass() {
					return eventClass;
				}

				public Collection<XAttribute> getAttributes() {
					return attributes;
				}
			};
		}
	}

	private final static class MapVariableProvider implements VariableProvider {

		private final Map<String, Object> attributes;

		private MapVariableProvider(Map<String, Object> attributes) {
			this.attributes = attributes;
		}

		public Object getValue(String variableName) throws VariableNotFoundException {
			return attributes.get(variableName);
		}

		public Map<String, Object> getAttributes() {
			return attributes;
		}

	}

	private final EventRelationStorage eventStore;

	// stores information the selected attributes only
	private ImmutableSet<String> consideredAttributes = ImmutableSet.of();

	// stores information about all attributes
	private final ImmutableMap<String, Class<?>> availableAttributes;
	private final ImmutableMap<String, Type> attributeTypesForDiscovery;

	private final Map<XEventClass, Multiset<Map<String, Object>>> occursWithAttributes = new HashMap<>();
	private final Map<Relation, Multiset<Map<String, Object>>> directlyFollowsUnderAttributes = new HashMap<>();
	private final Map<Relation, Multiset<Map<String, Object>>> lengthTwoLoopUnderAttributes = new HashMap<>();
	private final Map<String, Set<String>> literalValues = new HashMap<>();

	public DataRelationStorageImpl(EventRelationStorage eventRelationStorage, Map<String, Class<?>> availableAttributes,
			Set<String> consideredAttributes) {
		this.eventStore = eventRelationStorage;
		this.availableAttributes = ImmutableSortedMap.copyOf(availableAttributes);
		this.attributeTypesForDiscovery = ImmutableMap.copyOf(getAttributeTypeMap(availableAttributes));
		for (java.util.Map.Entry<String, Type> entry : attributeTypesForDiscovery.entrySet()) { // normal attribute key
			if (entry.getValue() == Type.LITERAL) {
				literalValues.put(entry.getKey(), Sets.<String>newConcurrentHashSet());
			}
		}
		for (Relation relation : eventRelationStorage.allCombinations()) {
			if (eventRelationStorage.countDirectlyFollows(relation) > 0) {
				directlyFollowsUnderAttributes.put(relation, ConcurrentHashMultiset.<Map<String, Object>>create());
			}
			if (eventRelationStorage.countLengthTwoFollows(relation.getSource(), relation.getTarget()) > 0) {
				lengthTwoLoopUnderAttributes.put(relation, ConcurrentHashMultiset.<Map<String, Object>>create());
			}
		}
		for (XEventClass eventClass : eventRelationStorage.getEventClasses()) {
			occursWithAttributes.put(eventClass, ConcurrentHashMultiset.<Map<String, Object>>create());
		}
		this.consideredAttributes = ImmutableSet.copyOf(consideredAttributes);
	}

	private static Map<String, Type> getAttributeTypeMap(Map<String, Class<?>> attributes) {
		Map<String, Type> attributeTypeMap = new HashMap<>();
		for (java.util.Map.Entry<String, Class<?>> entry : attributes.entrySet()) {
			Class<?> type = entry.getValue();
			String key = WekaUtil.fixVarName(entry.getKey());
			if (type == Boolean.class) {
				attributeTypeMap.put(key, Type.BOOLEAN);
			} else if (type == Double.class || type == Float.class) {
				attributeTypeMap.put(key, Type.CONTINUOS);
			} else if (type == Long.class || type == Integer.class) {
				attributeTypeMap.put(key, Type.DISCRETE);
			} else if (type == Date.class) {
				attributeTypeMap.put(key, Type.TIMESTAMP);
			} else if (type == String.class) {
				attributeTypeMap.put(key, Type.LITERAL);
			} else {
				throw new IllegalArgumentException("Does not support " + type);
			}
		}
		return attributeTypeMap;
	}

	public ImmutableSet<String> getConsideredAttributes() {
		return consideredAttributes;
	}

	public void setConsideredAttributes(final Set<String> consideredAttributes, ExecutorService executor) {
		if (!consideredAttributes.equals(this.consideredAttributes)) {
			this.consideredAttributes = ImmutableSet.copyOf(consideredAttributes);
			// clear relation cache
			for (Multiset<Map<String, Object>> ocSet : occursWithAttributes.values()) {
				ocSet.clear();
			}
			for (Multiset<Map<String, Object>> dfSet : directlyFollowsUnderAttributes.values()) {
				dfSet.clear();
			}
			for (Multiset<Map<String, Object>> tfSet : lengthTwoLoopUnderAttributes.values()) {
				tfSet.clear();
			}

			// Always register to build the literal lookup cache for the query helper
			try {
				TaskExecutor.Factory.getInstance().execute(executor, eventStore.getAllTraces(), new Command<XTrace>() {

					public void execute(XTrace trace) {
						registerTrace(trace);

					}
				});
			} catch (InterruptedException e) {
				Thread.currentThread().interrupt();
			}
		}

	}

	public void setConsideredAttributes(final Set<String> consideredAttributes) {
		ExecutorService executor = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
		try {
			setConsideredAttributes(consideredAttributes, executor);
		} finally {
			executor.shutdown();
		}
	}

	public void registerTrace(XTrace trace) {

		// Create an iterator with artificial start/end events
		PeekingIterator<AttributedEvent> eventIter = createEventIterator(trace);
		XEventClass lastEventClass = null;

		Map<String, Object> attributes = new HashMap<>();
		ImmutableMap<String, Object> lastAttributes = ImmutableMap.of();

		while (eventIter.hasNext()) {
			AttributedEvent currentEvent = eventIter.next();
			XEventClass currentClass = currentEvent.getEventClass();

			for (XAttribute attr : currentEvent.getAttributes()) {
				String key = attr.getKey();
				if (availableAttributes.containsKey(key)) {
					String varName = WekaUtil.fixVarName(key);
					if (consideredAttributes.contains(key)) {
						Object val = XUtils.getAttributeValue(attr);
						if (attr instanceof XAttributeLiteral) {
							literalValues.get(varName).add((String) val);
						}
						attributes.put(varName, val);
					} else {
						if (attr instanceof XAttributeLiteral) {
							literalValues.get(varName).add((String) XUtils.getAttributeValue(attr));
						}
					}
				}
			}

			ImmutableMap<String, Object> currentAttributes = ImmutableMap.copyOf(attributes);

			occursWithAttributes.get(currentClass).add(currentAttributes);

			if (eventIter.hasNext()) {
				AttributedEvent followingEvent = eventIter.peek();
				XEventClass followingClass = followingEvent.getEventClass();

				// ab -> b [directly follows] a
				Relation dfRelation = Relation.Factory.create(currentClass, followingClass);
				directlyFollowsUnderAttributes.get(dfRelation).add(currentAttributes);

				// aba -> b [in length two loop with] a
				if (followingClass.equals(lastEventClass)) {
					lengthTwoLoopUnderAttributes.get(Relation.Factory.create(lastEventClass, currentClass))
							.add(lastAttributes);
				}
			}

			lastAttributes = ImmutableMap.copyOf(attributes);
			lastEventClass = currentClass;
		}
	}

	private PeekingIterator<AttributedEvent> createEventIterator(XTrace trace) {
		final Map<String, XAttribute> attributeValues = trace.getAttributes();

		// Add initial and last event classes
		AttributedEvent startEvent = new AttributedEvent() {

			public XEventClass getEventClass() {
				return eventStore.getStartEventClass();
			}

			public Collection<XAttribute> getAttributes() {
				return attributeValues.values();
			}
		};
		AttributedEvent endEvent = new AttributedEvent() {

			public XEventClass getEventClass() {
				return eventStore.getEndEventClass();
			}

			public Collection<XAttribute> getAttributes() {
				return ImmutableList.of(); // no new attributes
			}
		};

		return Iterators.peekingIterator(Iterators.concat(Iterators.singletonIterator(startEvent),
				Iterators.transform(trace.iterator(), new AttributedEventTransformation()),
				Iterators.singletonIterator(endEvent)));
	}

	@Override
	public DependencyRelation getDependencyRelationUnderCondition(Relation followsRelation, GuardExpression condition) {
		XEventClass source = followsRelation.getSource();
		XEventClass target = followsRelation.getTarget();
		int sourceToTarget = countDirectlyFollows(source, target, condition);
		int targetToSource = countDirectlyFollows(target, source, condition);
		final double dependencyValue = (sourceToTarget - targetToSource)
				/ (double) (sourceToTarget + targetToSource + 1);
		return new DependencyRelationImpl(dependencyValue);
	}

	@Override
	public DependencyRelation getL1RelationUnderCondition(Relation relation, GuardExpression condition) {
		XEventClass source = relation.getSource();
		XEventClass target = relation.getTarget();
		final double dependencyValue = (countDirectlyFollows(source, target, condition))
				/ (double) (countDirectlyFollows(source, target, condition) + 1);
		return new DependencyRelationImpl(dependencyValue);
	}

	@Override
	public DependencyRelation getL2RelationUnderCondition(Relation relation, GuardExpression condition) {
		XEventClass source = relation.getSource();
		XEventClass target = relation.getTarget();
		final double dependencyValue = (countLengthTwoFollows(source, target, condition)
				+ countLengthTwoFollows(target, source, condition))
				/ (double) (countLengthTwoFollows(source, target, condition)
						+ countLengthTwoFollows(target, source, condition) + 1);
		return new DependencyRelationImpl(dependencyValue);
	}

	@Override
	public int countOccurence(XEventClass eventClass, GuardExpression condition) {
		if (eventStore.countOccurence(eventClass) > 0) {

			Set<String> attributeInCondition = condition.getNormalVariables();

			Multiset<Map<String, Object>> observedAttributes = occursWithAttributes.get(eventClass);

			int occursUnderCondition = 0;
			for (Entry<Map<String, Object>> instance : observedAttributes.entrySet()) {
				final MapVariableProvider variableProvider = new MapVariableProvider(instance.getElement());
				if (conditionHolds(condition, variableProvider, attributeInCondition)) {
					occursUnderCondition += instance.getCount();
				}
			}

			return occursUnderCondition;

		} else {

			// Cannot be filtered more by a condition
			return 0;
		}
	}

	@Override
	public int countDirectlyFollows(XEventClass source, XEventClass target, GuardExpression condition) {
		return countDirectlyFollows(Relation.Factory.create(source, target), condition);
	}

	public Multiset<Map<String, Object>> directlyFollowsUnderAttributes(Relation relation) {
		return directlyFollowsUnderAttributes.get(relation);
	}

	@Override
	public int countDirectlyFollows(Relation relation, GuardExpression condition) {
		if (eventStore.countDirectlyFollows(relation) > 0) {

			Set<String> attributeInCondition = condition.getNormalVariables();

			Multiset<Map<String, Object>> observedAttributes = directlyFollowsUnderAttributes.get(relation);

			int followsUnderCondition = 0;
			for (Entry<Map<String, Object>> instance : observedAttributes.entrySet()) {
				final MapVariableProvider variableProvider = new MapVariableProvider(instance.getElement());
				if (conditionHolds(condition, variableProvider, attributeInCondition)) {
					followsUnderCondition += instance.getCount();
				}
			}

			return followsUnderCondition;

		} else {

			// Cannot be filtered more by a condition
			return 0;
		}
	}

	private static boolean conditionHolds(GuardExpression condition, final MapVariableProvider variableProvider,
			Set<String> attributeInCondition) {
		if (variableProvider.getAttributes().keySet().containsAll(attributeInCondition)) {
			return condition.isTrue(variableProvider);
		} else {
			return false;
		}
	}

	public int countLengthTwoFollows(Relation relation, GuardExpression condition) {
		return countLengthTwoFollows(relation.getSource(), relation.getTarget(), condition);
	}

	public int countLengthTwoFollows(XEventClass source, XEventClass target, GuardExpression condition) {
		if (eventStore.countLengthTwoFollows(source, target) > 0) {

			Set<String> attributeInCondition = condition.getNormalVariables();

			Multiset<Map<String, Object>> observedAttributes = lengthTwoLoopUnderAttributes
					.get(Relation.Factory.create(source, target));

			int followsUnderCondition = 0;
			for (Entry<Map<String, Object>> instance : observedAttributes.entrySet()) {
				final MapVariableProvider variableProvider = new MapVariableProvider(instance.getElement());
				if (conditionHolds(condition, variableProvider, attributeInCondition)) {
					followsUnderCondition += instance.getCount();
				}
			}

			return followsUnderCondition;

		} else {

			// Cannot be filtered more by a condition
			return 0;
		}
	}

	public ImmutableMap<String, Type> getAttributeTypes() {
		return attributeTypesForDiscovery;
	}

	public Map<String, Set<String>> getLiteralValues() {
		return literalValues;
	}

	public ImmutableMap<String, Class<?>> getAvailableAttributes() {
		return availableAttributes;
	}

}