package org.processmining.dataawarecnetminer.plugins;

import java.text.MessageFormat;

import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.models.cnet.CNet;
import org.processmining.models.cnet.CNetBinding;
import org.processmining.models.cnet.CNetNode;
import org.processmining.models.cnet.CausalNet;
import org.processmining.models.graphbased.directed.DirectedGraph;
import org.processmining.models.graphbased.directed.DirectedGraphEdge;
import org.processmining.models.graphbased.directed.DirectedGraphNode;
import org.processmining.models.visualcnet.CNetV;
import org.processmining.petrinets.analysis.gedsim.algorithms.GraphEditDistanceSimilarityAlgorithm;
import org.processmining.petrinets.analysis.gedsim.algorithms.impl.GraphEditDistanceSimilarityAStar;
import org.processmining.petrinets.analysis.gedsim.params.GraphEditDistanceSimilarityParameters;
import org.processmining.plugins.utils.ProvidedObjectHelper;

/**
 * Computes the GED of two Causal Nets by considering the edit distance of the
 * dependency graph only. Differences in the bindings are not considered. Uses
 * the A* algorithm described in:
 * 
 * <pre>
 * Dijkman, Remco, Marlon Dumas, and Luciano Garcia-Banuelos. 
 * "Graph matching algorithms for business process model similarity search." 
 * International Conference on Business Process Management. 
 * Springer Berlin Heidelberg, 2009.
 * </pre>
 * 
 * @author F. Mannhardt
 *
 */
public final class CausalNetGEDComparison {

	@Plugin(name = "Compare Causal Nets (GED)", parameterLabels = { "Causal Net - A", "Causal Net - B" }, //
			level = PluginLevel.Regular, returnLabels = { "Graph Edit Distance" }, returnTypes = {
					String.class }, userAccessible = true, mostSignificantResult = 1, help = "Computes the GED of two Causal Nets by considering the edit distance of the dependency graph only. Differences in the bindings are not considered. Uses the A* algorithm described in: \r\n"
							+ " Dijkman, Remco, Marlon Dumas, and Luciano Garcia-Banuelos. \r\n"
							+ " \"Graph matching algorithms for business process model similarity search.\" \r\n"
							+ " International Conference on Business Process Management. \r\n"
							+ " Springer Berlin Heidelberg, 2009.\r\n")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")

	public String graphEditDistanceCausalNet(PluginContext context, CausalNet causalNetA, CausalNet causalNetB) {

		double distanceAStar = graphEditDistance(causalNetA, causalNetB);

		String labelA = ProvidedObjectHelper.getProvidedObjectLabel(context, causalNetA);
		String labelB = ProvidedObjectHelper.getProvidedObjectLabel(context, causalNetB);

		String result = MessageFormat.format("GED-Distance between {0} and {1} (A*): {2}", labelA, labelB,
				distanceAStar);
		System.out.println(result);
		String htmlResult = MessageFormat.format("<html>{0}</html>", result);
		context.getFutureResult(0).setLabel(result);
		return htmlResult;
	}

	public static double graphEditDistance(CausalNet causalNetA, CausalNet causalNetB) {
		CNet cnetA = convertToCNet(causalNetA);
		CNet cnetB = convertToCNet(causalNetB);

		DirectedGraph<? extends DirectedGraphNode, ? extends DirectedGraphEdge<?, ?>> graph1 = new CNetV(cnetA);
		DirectedGraph<? extends DirectedGraphNode, ? extends DirectedGraphEdge<?, ?>> graph2 = new CNetV(cnetB);

		GraphEditDistanceSimilarityParameters parameters = new GraphEditDistanceSimilarityParameters();
		parameters.setUsePureDistance(true);
		parameters.setLedCutOff(0.0);

		GraphEditDistanceSimilarityAlgorithm<DirectedGraph<? extends DirectedGraphNode, ? extends DirectedGraphEdge<?, ?>>> aStarGed = new GraphEditDistanceSimilarityAStar<>(
				parameters);
		double distanceAStar = aStarGed.compute(graph1, graph2);
		return distanceAStar;
	}

	private static CNet convertToCNet(CausalNet causalNetA) {
		CNet cnet = new CNet(causalNetA.getLabel());
		for (CNetNode node : causalNetA.getNodes()) {
			cnet.addNode(node);
			for (CNetBinding binding : causalNetA.getOutputBindings(node)) {
				cnet.addOutputBinding(node, binding.getBoundNodes());
			}
			for (CNetBinding binding : causalNetA.getInputBindings(node)) {
				cnet.addInputBinding(node, binding.getBoundNodes());
			}
		}
		return cnet;
	}

}