package org.processmining.dataawarecnetminer.visualization;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dialog.ModalityType;
import java.awt.event.ActionEvent;
import java.text.MessageFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTable;
import javax.swing.RowSorter;
import javax.swing.SortOrder;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableModel;
import javax.swing.table.TableRowSorter;

import org.processmining.dataawarecnetminer.model.FrequencyAwareCausalNet;
import org.processmining.dataawarecnetminer.model.FrequencyAwareDataCausalNet;
import org.processmining.dataawarecnetminer.model.FrequencyAwareNet;
import org.processmining.framework.util.ui.widgets.ProMTableWithoutPanel;
import org.processmining.models.cnet.CNetBinding;
import org.processmining.models.cnet.CNetBinding.Type;
import org.processmining.models.cnet.CNetNode;
import org.processmining.models.cnet.CausalNet;
import org.processmining.plugins.dcnet.DefaultNodeDecorator;

import com.google.common.base.Joiner;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Ordering;

public class FrequencyNodeDecorator extends DefaultNodeDecorator {

	private static final Color GOOD_MOVE_COLOR = Color.GREEN;
	private static final Color MODEL_MODE_COLOR = new Color(224, 176, 255);
	private static final Color LOG_MOVE_COLOR = Color.YELLOW;

	private final class TrafficLightSegmentAlignmentImpl implements TrafficLightSegment {

		private final Color color;

		private TrafficLightSegmentAlignmentImpl(Color color) {
			super();
			this.color = color;
		}

		public Color getColor() {
			return color;
		}

		public double getSize() {
			return 2;
		}
	}

	private final static class ColorSegmentImpl implements ColorSegment {

		private double size;
		private Color color;

		public ColorSegmentImpl(double size, Color color) {
			this.size = size;
			this.color = color;
		}

		public double getSize() {
			return size;
		}

		public Color getColor() {
			return color;
		}
	}

	private static final class ShowBindingTableAction extends AbstractAction {

		private static final long serialVersionUID = 1L;

		private final JPopupMenu menu;

		private final FrequencyAwareNet frequencyAwareNet;
		private final CausalNet causalNet;
		private final CNetNode node;
		private final Type bindingType;

		public ShowBindingTableAction(String name, Type bindingType, FrequencyAwareNet frequencyAwareNet,
				CausalNet causalNet, CNetNode node, JPopupMenu menu) {
			super(name);
			this.bindingType = bindingType;
			this.node = node;
			this.menu = menu;
			this.frequencyAwareNet = frequencyAwareNet;
			this.causalNet = causalNet;
		}

		public void actionPerformed(ActionEvent e) {
			DefaultTableModel tableModel = createTableModel();
			String dialogTitle;
			if (bindingType == Type.INPUT) {
				for (CNetBinding binding : causalNet.getInputBindings(node)) {
					addBindingToTable(tableModel, binding, frequencyAwareNet.getBindingFrequency(binding),
							frequencyAwareNet.getInputFrequency(node));
				}
				dialogTitle = "Input bindings for " + node.getLabel();
			} else {
				for (CNetBinding binding : causalNet.getOutputBindings(node)) {
					addBindingToTable(tableModel, binding, frequencyAwareNet.getBindingFrequency(binding),
							frequencyAwareNet.getOutputFrequency(node));
				}
				dialogTitle = "Output bindings for " + node.getLabel();
			}
			ProMTableWithoutPanel table = createTable(tableModel);
			sortByFrequency(table);
			UIUtil.showCustom(menu, table, dialogTitle, ModalityType.MODELESS);
		}

		private DefaultTableModel createTableModel() {
			DefaultTableModel tableModel = new DefaultTableModel(
					new String[] { "Binding", "Absolute Frequency", "Relative Frequency" }, 0) {

				private static final long serialVersionUID = 1L;

				public Class<?> getColumnClass(int columnIndex) {
					switch (columnIndex) {
						case 0 :
							return String.class;
						case 1 :
							return Integer.class;
						case 2 :
							return Double.class;
						default :
							return String.class;
					}
				}

			};
			return tableModel;
		}

		private ProMTableWithoutPanel createTable(DefaultTableModel tableModel) {
			ProMTableWithoutPanel table = new ProMTableWithoutPanel(tableModel);
			table.setAutoResizeMode(JTable.AUTO_RESIZE_ALL_COLUMNS);
			table.setAutoCreateRowSorter(true);
			table.getColumnModel().getColumn(1).setCellRenderer(new DefaultTableCellRenderer() {

				private static final long serialVersionUID = 1L;

				public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected,
						boolean hasFocus, int row, int column) {
					return super.getTableCellRendererComponent(table, NumberFormat.getIntegerInstance().format(value),
							isSelected, hasFocus, row, column);
				}

			});
			table.getColumnModel().getColumn(2).setCellRenderer(new DefaultTableCellRenderer() {

				private static final long serialVersionUID = 1L;

				public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected,
						boolean hasFocus, int row, int column) {
					return super.getTableCellRendererComponent(table, NumberFormat.getPercentInstance().format(value),
							isSelected, hasFocus, row, column);
				}

			});
			return table;
		}

		private void sortByFrequency(ProMTableWithoutPanel table) {
			TableRowSorter<TableModel> sorter = new TableRowSorter<>(table.getModel());
			table.setRowSorter(sorter);
			sorter.setSortKeys(ImmutableList.of(new RowSorter.SortKey(1, SortOrder.DESCENDING)));
			sorter.sort();
		}

		private void addBindingToTable(DefaultTableModel tableModel, CNetBinding binding, int bindingFrequency,
				int totalLocalFrequency) {
			Iterable<CNetNode> sortedBinding = Ordering.natural().sortedCopy(binding.getBoundNodes());
			String formattedBinding = Joiner.on(", ").join(sortedBinding);
			tableModel.addRow(new Object[] { formattedBinding, Integer.valueOf(bindingFrequency),
					Double.valueOf(bindingFrequency / (double) totalLocalFrequency) });
		}
	}

	private final FrequencyAwareNet frequencyAwareNet;
	private final CausalNet causalNet;

	private TransformationMode transformationMode = TransformationMode.LINEAR;
	private NumberFormat frequencyFormatter = NumberFormat.getIntegerInstance();
	private double minBindingSize = 4.0;
	private double maxBindingSize = 8.0;
	private FrequencyMode frequencyMode = FrequencyMode.GLOBAL;

	private final int maxFrequency;
	private ColorScheme colorScheme = ColorScheme.BLUE_SINGLE_HUE;

	public FrequencyNodeDecorator(FrequencyAwareCausalNet frequencyCausalNet) {
		super(frequencyCausalNet.getCNet());
		this.frequencyAwareNet = frequencyCausalNet;
		this.causalNet = frequencyCausalNet.getCNet();
		this.maxFrequency = calcMaxFrequency(frequencyCausalNet.getCNet(), frequencyCausalNet);
	}

	public FrequencyNodeDecorator(FrequencyAwareDataCausalNet frequencyDataCausalNet) {
		super(frequencyDataCausalNet.getDCNet());
		this.frequencyAwareNet = frequencyDataCausalNet;
		this.causalNet = frequencyDataCausalNet.getDCNet().getCausalNet();
		this.maxFrequency = calcMaxFrequency(frequencyDataCausalNet.getDCNet().getCausalNet(), frequencyDataCausalNet);
	}

	private static int calcMaxFrequency(CausalNet cnet, FrequencyAwareNet frequencyCausalNet) {
		int frequency = 0;
		for (CNetNode node : cnet.getNodes()) {
			frequency = Math.max(frequency, frequencyCausalNet.getOutputFrequency(node));
		}
		return frequency;
	}

	public String getNodeLabel(CNetNode node) {
		return frequencyFormatter.format(frequencyAwareNet.getNodeFrequency(node));
	}

	public JPopupMenu getNodeMenu(CNetNode node) {
		JPopupMenu menu = super.getNodeMenu(node);
		menu.add(new ShowBindingTableAction("Show input bindings", CNetBinding.Type.INPUT, frequencyAwareNet, causalNet,
				node, menu));
		menu.add(new ShowBindingTableAction("Show output bindings", CNetBinding.Type.OUTPUT, frequencyAwareNet,
				causalNet, node, menu));
		int skippedFrequency = frequencyAwareNet.getNodeSkippedFrequency(node);
		int insertedFrequency = frequencyAwareNet.getNodeInsertedFrequency(node);
		if (skippedFrequency > 0 || insertedFrequency > 0) {
			menu.addSeparator();
		}
		if (skippedFrequency > 0) {
			menu.add(new JMenuItem("Skipped (log move) " + skippedFrequency + " times"));
		}
		if (insertedFrequency > 0) {
			menu.add(new JMenuItem("Inserted (model move) " + insertedFrequency + " times "));
		}
		if (frequencyAwareNet instanceof FrequencyAwareDataCausalNet) {
			for (String variable : ((FrequencyAwareDataCausalNet) frequencyAwareNet).getDCNet().getWrites(node)) {
				int correct = frequencyAwareNet.getCorrectVariableFrequency(node, variable);
				int incorrect = frequencyAwareNet.getIncorrectVariableFrequency(node, variable);
				int missing = frequencyAwareNet.getMissingVariableFrequency(node, variable);
				if (incorrect > 0 || missing > 0) {
					menu.add(new JMenuItem(
							MessageFormat.format("Writes variable \"{0}\": {1} correct, {2} incorrect, {3} missing",
									variable, correct, incorrect, missing)));
				} else {
					menu.add(new JMenuItem(MessageFormat.format("Writes variable \"{0}\"", variable)));
				}
			}
		}
		return menu;
	}

	public List<ColorSegment> getNodeColorSegments(CNetNode node) {
		int nodeFrequency = frequencyAwareNet.getNodeFrequency(node);
		int skippedFrequency = frequencyAwareNet.getNodeSkippedFrequency(node);
		int insertedFrequency = frequencyAwareNet.getNodeInsertedFrequency(node);
		if (skippedFrequency > 0 || insertedFrequency > 0) {
			ArrayList<ColorSegment> segments = new ArrayList<>();
			segments.add(new ColorSegmentImpl(skippedFrequency / (double) nodeFrequency, LOG_MOVE_COLOR));
			segments.add(new ColorSegmentImpl(insertedFrequency / (double) nodeFrequency, MODEL_MODE_COLOR));
			segments.add(new ColorSegmentImpl(
					(nodeFrequency - skippedFrequency - insertedFrequency) / (double) nodeFrequency, GOOD_MOVE_COLOR));
			return segments;
		} else {
			return super.getNodeColorSegments(node);
		}
	}

	public List<TrafficLightSegment> getNodeTrafficLights(CNetNode node) {
		int skippedFrequency = frequencyAwareNet.getNodeSkippedFrequency(node);
		int insertedFrequency = frequencyAwareNet.getNodeInsertedFrequency(node);
		if (skippedFrequency > 0 || insertedFrequency > 0) {
			ArrayList<TrafficLightSegment> segments = new ArrayList<>();
			if (skippedFrequency > 0) {
				segments.add(new TrafficLightSegmentAlignmentImpl(LOG_MOVE_COLOR));
			}
			if (insertedFrequency > 0) {
				segments.add(new TrafficLightSegmentAlignmentImpl(MODEL_MODE_COLOR));
			}
			return segments;
		} else {
			return super.getNodeTrafficLights(node);
		}
	}

	public String getBindingLabel(CNetBinding binding) {
		return frequencyFormatter.format(frequencyAwareNet.getBindingFrequency(binding));
	}

	public double getBindingSize(CNetBinding binding) {
		double frequencyRatio = calcFrequencyRatio(binding);
		if (frequencyRatio == 0.0) {
			return (float) (getMinBindingSize() / 2.0);
		}
		return getMinBindingSize() + frequencyRatio * (getMaxBindingSize() - getMinBindingSize());
	}

	public Color getBindingFillColor(CNetBinding binding) {
		float frequency = Math.min(1.0f, (float) calcFrequencyRatio(binding));
		if (frequency == 0.0) {
			return Color.LIGHT_GRAY;
		}
		return colorScheme.getColorFromGradient(frequency);
	}

	public Color getBindingBorderColor(CNetBinding binding) {
		float frequency = Math.min(1.0f, (float) calcFrequencyRatio(binding));
		if (frequency == 0.0) {
			return Color.LIGHT_GRAY;
		}
		return colorScheme.getColorFromGradient(frequency);
	}

	public Color getNodeFillColor(CNetNode node) {
		double frequency = calcFrequencyRatio(node);
		if (frequency == 0.0) {
			return Color.LIGHT_GRAY;
		}
		return colorScheme.getColorFromGradient(frequency);
	}

	public Color getBindingEdgeColor(CNetBinding binding) {
		float frequency = Math.min(1.0f, (float) calcFrequencyRatio(binding));
		if (frequency == 0.0) {
			return Color.LIGHT_GRAY;
		}
		return colorScheme.getColorFromGradient(frequency);
	}

	public float getBindingEdgeWidth(CNetBinding binding) {
		double frequencyRatio = calcFrequencyRatio(binding);
		return (float) (getMinBindingSize() / 4.0
				+ frequencyRatio * ((getMaxBindingSize() - getMinBindingSize()) / 4.0));
	}

	private double calcFrequencyRatio(CNetNode node) {
		int frequency = frequencyAwareNet.getNodeFrequency(node);
		return transformationMode.transform(maxFrequency, frequency);
	}

	private double calcFrequencyRatio(CNetBinding binding) {
		int totalFrequency = 0;
		if (frequencyMode == FrequencyMode.GLOBAL) {
			totalFrequency = maxFrequency;
		} else {
			if (binding.getType() == Type.INPUT) {
				totalFrequency = frequencyAwareNet.getInputFrequency(binding.getNode());
			} else {
				totalFrequency = frequencyAwareNet.getOutputFrequency(binding.getNode());
			}
		}
		int frequency = frequencyAwareNet.getBindingFrequency(binding);
		double frequencyRatio = transformationMode.transform(totalFrequency, frequency);
		return frequencyRatio;
	}

	public TransformationMode getTransformationMode() {
		return transformationMode;
	}

	public void setTransformationMode(TransformationMode transformationMode) {
		this.transformationMode = transformationMode;
	}

	public NumberFormat getFrequencyFormatter() {
		return frequencyFormatter;
	}

	public void setFrequencyFormatter(NumberFormat frequencyFormatter) {
		this.frequencyFormatter = frequencyFormatter;
	}

	public double getMinBindingSize() {
		return minBindingSize;
	}

	public void setMinBindingSize(double minBindingSize) {
		this.minBindingSize = minBindingSize;
	}

	public double getMaxBindingSize() {
		return maxBindingSize;
	}

	public void setMaxBindingSize(double maxBindingSize) {
		this.maxBindingSize = maxBindingSize;
	}

}